from .pool import deferToAMPProcess, pp
from .commands import Shutdown, Ping, Echo
from .child import AMPChild
from ._version import __version__ as _my_version

__version__ = _my_version.short()


__all__ = [
    'deferToAMPProcess',
    'pp',
    'Shutdown', 'Ping', 'Echo',
    'AMPChild',
    '__version__',
]



from twisted.protocols import amp
from twisted.python import reflect
from twisted.internet import defer

from ampoule import child

class Function(amp.Argument):
    def toString(self, inObject):
        return reflect.qual(inObject)

    def fromString(self, inString):
        return reflect.namedAny(inString)


class Map(amp.Command):
    arguments = [('mapper', Function()),
                 ('filename', amp.Path()),
                 ('outdir', amp.Path())]

    response = [('result', amp.Path())]

class Reduce(amp.Command):
    arguments = [('reducer', Function()),
                 ('directory', amp.Path())]

    response = [('result', amp.Path())]

class MapReducer(child.AMPChild):
    def _call(self, fun, in_, out):
        return defer.maybeDeferred(fun, in_, out
            ).addCallback(lambda _: {'result': out})

    @Map.responder
    def map(self, mapper, filename, outdir):
        in_ = filename
        out = outdir.child(filename.basename()).siblingExtension('.map')
        return self._call(mapper, in_, out)

    @Reduce.responder
    def reduce(self, reducer, directory):
        in_ = directory.globChildren('*.map')
        out = directory.child('reduced.red')
        return self._call(reducer, in_, out)



import os
import sys
import imp
import itertools

from zope.interface import implementer

from twisted import logger
from twisted.internet import reactor, protocol, defer, error
from twisted.python import reflect
from twisted.protocols import amp
from twisted.python import runtime
from twisted.python.compat import set

from ampoule import iampoule



log = logger.Logger()


gen = itertools.count()

if runtime.platform.isWindows():
    IS_WINDOWS = True
    TO_CHILD = 0
    FROM_CHILD = 1
else:
    IS_WINDOWS = False
    TO_CHILD = 3
    FROM_CHILD = 4

class AMPConnector(protocol.ProcessProtocol):
    """
    A L{ProcessProtocol} subclass that can understand and speak AMP.

    @ivar amp: the children AMP process
    @type amp: L{amp.AMP}

    @ivar finished: a deferred triggered when the process dies.
    @type finished: L{defer.Deferred}

    @ivar name: Unique name for the connector, much like a pid.
    @type name: int
    """

    def __init__(self, proto, name=None):
        """
        @param proto: An instance or subclass of L{amp.AMP}
        @type proto: L{amp.AMP}

        @param name: optional name of the subprocess.
        @type name: int
        """
        self.finished = defer.Deferred()
        self.amp = proto
        self.name = name
        if name is None:
            self.name = next(gen)

    def signalProcess(self, signalID):
        """
        Send the signal signalID to the child process

        @param signalID: The signal ID that you want to send to the
                        corresponding child
        @type signalID: C{str} or C{int}
        """
        return self.transport.signalProcess(signalID)

    def connectionMade(self):
        log.info(u'Subprocess {n} started.', n=self.name)
        self.amp.makeConnection(self)

    # Transport
    disconnecting = False

    def write(self, data):
        if IS_WINDOWS:
            self.transport.write(data)
        else:
            self.transport.writeToChild(TO_CHILD, data)

    def loseConnection(self):
        self.transport.closeChildFD(TO_CHILD)
        self.transport.closeChildFD(FROM_CHILD)
        self.transport.loseConnection()

    def getPeer(self):
        return ('subprocess',)

    def getHost(self):
        return ('no host',)

    def childDataReceived(self, childFD, data):
        if childFD == FROM_CHILD:
            self.amp.dataReceived(data)
            return
        self.errReceived(data)

    def errReceived(self, data):
        for line in data.strip().splitlines():
            log.error(u'FROM {n}: {l}', n=self.name, l=line)

    def processEnded(self, status):
        log.info(u'Process: {n} ended', n=self.name)
        self.amp.connectionLost(status)
        if status.check(error.ProcessDone):
            self.finished.callback('')
            return
        self.finished.errback(status)

BOOTSTRAP = """\
import sys

def main(reactor, ampChildPath):
    from twisted.application import reactors
    reactors.installReactor(reactor)

    from twisted import logger
    observer = logger.textFileLogObserver(sys.stderr)
    logLevelPredicate = logger.LogLevelFilterPredicate(
        defaultLogLevel=logger.LogLevel.info
    )
    filteringObserver = logger.FilteringLogObserver(
        observer, [logLevelPredicate]
    )
    logger.globalLogBeginner.beginLoggingTo([filteringObserver])

    from twisted.internet import reactor, stdio
    from twisted.python import reflect, runtime

    ampChild = reflect.namedAny(ampChildPath)
    ampChildInstance = ampChild(*sys.argv[1:-2])
    if runtime.platform.isWindows():
        stdio.StandardIO(ampChildInstance)
    else:
        stdio.StandardIO(ampChildInstance, %s, %s)
    enter = getattr(ampChildInstance, '__enter__', None)
    if enter is not None:
        enter()
    try:
        reactor.run()
    except:
        if enter is not None:
            info = sys.exc_info()
            if not ampChildInstance.__exit__(*info):
                raise
        else:
            raise
    else:
        if enter is not None:
            ampChildInstance.__exit__(None, None, None)

main(sys.argv[-2], sys.argv[-1])
""" % (TO_CHILD, FROM_CHILD)

@implementer(iampoule.IStarter)
class ProcessStarter(object):

    connectorFactory = AMPConnector
    def __init__(self, bootstrap=BOOTSTRAP, args=(), env={},
                 path=None, uid=None, gid=None, usePTY=0,
                 packages=(), childReactor="select"):
        """
        @param bootstrap: Startup code for the child process
        @type  bootstrap: C{str}

        @param args: Arguments that should be supplied to every child
                     created.
        @type args: C{tuple} of C{str}

        @param env: Environment variables that should be present in the
                    child environment
        @type env: C{dict}

        @param path: Path in which to run the child
        @type path: C{str}

        @param uid: if defined, the uid used to run the new process.
        @type uid: C{int}

        @param gid: if defined, the gid used to run the new process.
        @type gid: C{int}

        @param usePTY: Should the child processes use PTY processes
        @type usePTY: 0 or 1

        @param packages: A tuple of packages that should be guaranteed
                         to be importable in the child processes
        @type packages: C{tuple} of C{str}

        @param childReactor: a string that sets the reactor for child
                             processes
        @type childReactor: C{str}
        """
        self.bootstrap = bootstrap
        self.args = args
        self.env = env
        self.path = path
        self.uid = uid
        self.gid = gid
        self.usePTY = usePTY
        self.packages = ("ampoule",) + packages
        self.childReactor = childReactor

    def __repr__(self):
        """
        Represent the ProcessStarter with a string.
        """
        return """ProcessStarter(bootstrap=%r,
                                 args=%r,
                                 env=%r,
                                 path=%r,
                                 uid=%r,
                                 gid=%r,
                                 usePTY=%r,
                                 packages=%r,
                                 childReactor=%r)""" % (self.bootstrap,
                                                        self.args,
                                                        self.env,
                                                        self.path,
                                                        self.uid,
                                                        self.gid,
                                                        self.usePTY,
                                                        self.packages,
                                                        self.childReactor)

    def _checkRoundTrip(self, obj):
        """
        Make sure that an object will properly round-trip through 'qual' and
        'namedAny'.

        Raise a L{RuntimeError} if they aren't.
        """
        tripped = reflect.namedAny(reflect.qual(obj))
        if tripped is not obj:
            raise RuntimeError("importing %r is not the same as %r" %
                               (reflect.qual(obj), obj))

    def startAMPProcess(self, ampChild, ampParent=None, ampChildArgs=()):
        """
        @param ampChild: a L{ampoule.child.AMPChild} subclass.
        @type ampChild: L{ampoule.child.AMPChild}

        @param ampParent: an L{amp.AMP} subclass that implements the parent
                          protocol for this process pool
        @type ampParent: L{amp.AMP}
        """
        self._checkRoundTrip(ampChild)
        fullPath = reflect.qual(ampChild)
        if ampParent is None:
            ampParent = amp.AMP
        prot = self.connectorFactory(ampParent())
        args = ampChildArgs + (self.childReactor, fullPath)
        return self.startPythonProcess(prot, *args)


    def startPythonProcess(self, prot, *args):
        """
        @param prot: a L{protocol.ProcessProtocol} subclass
        @type prot: L{protocol.ProcessProtocol}

        @param args: a tuple of arguments that will be added after the
                     ones in L{self.args} to start the child process.

        @return: a tuple of the child process and the deferred finished.
                 finished triggers when the subprocess dies for any reason.
        """
        spawnProcess(prot, self.bootstrap, self.args+args, env=self.env,
                     path=self.path, uid=self.uid, gid=self.gid,
                     usePTY=self.usePTY, packages=self.packages)

        # XXX: we could wait for startup here, but ... is there really any
        # reason to?  the pipe should be ready for writing.  The subprocess
        # might not start up properly, but then, a subprocess might shut down
        # at any point too. So we just return amp and have this piece to be
        # synchronous.
        return prot.amp, prot.finished

def spawnProcess(processProtocol, bootstrap, args=(), env={},
                 path=None, uid=None, gid=None, usePTY=0,
                 packages=()):
    env = env.copy()

    pythonpath = []
    for pkg in packages:
        p = os.path.split(imp.find_module(pkg)[1])[0]
        if p.startswith(os.path.join(sys.prefix, 'lib')):
            continue
        pythonpath.append(p)
    pythonpath = list(set(pythonpath))
    pythonpath.extend(env.get('PYTHONPATH', '').split(os.pathsep))
    env['PYTHONPATH'] = os.pathsep.join(pythonpath)
    args = (sys.executable, '-c', bootstrap) + args
    # childFDs variable is needed because sometimes child processes
    # misbehave and use stdout to output stuff that should really go
    # to stderr. Of course child process might even use the wrong FDs
    # that I'm using here, 3 and 4, so we are going to fix all these
    # issues when I add support for the configuration object that can
    # fix this stuff in a more configurable way.
    if IS_WINDOWS:
        return reactor.spawnProcess(processProtocol, sys.executable, args,
                                    env, path, uid, gid, usePTY)
    else:
        return reactor.spawnProcess(processProtocol, sys.executable, args,
                                    env, path, uid, gid, usePTY,
                                    childFDs={0:"w", 1:"r", 2:"r", 3:"w", 4:"r"})



import time
import random
import heapq
import itertools
import functools
import signal
choice = random.choice
now = time.time
count = functools.partial(next, itertools.count())
pop = heapq.heappop

from twisted import logger
from twisted.internet import defer, task, error
from twisted.python.failure import Failure

from ampoule import commands, main



log = logger.Logger()


STATS_TEMPLATE = u"""ProcessPool stats:
    workers:       {w}
    timeout:       {t}
    parent:        {p}
    child:         {c}
    max idle:      {i}
    recycle after: {r}
    ProcessStarter:
                   {s}"""


try:
    DIE = signal.SIGKILL
except AttributeError:
    # Windows doesn't have SIGKILL, let's just use SIGTERM then
    DIE = signal.SIGTERM


class ProcessPool(object):
    """
    This class generalizes the functionality of a pool of
    processes to which work can be dispatched.

    @ivar finished: Boolean flag, L{True} when the pool is finished.

    @ivar started: Boolean flag, L{True} when the pool is started.

    @ivar name: Optional name for the process pool

    @ivar min: Minimum number of subprocesses to set up

    @ivar max: Maximum number of subprocesses to set up

    @ivar maxIdle: Maximum number of seconds of indleness in a child

    @ivar starter: A process starter instance that provides
                    L{iampoule.IStarter}.

    @ivar recycleAfter: Maximum number of calls before restarting a
                        subprocess, 0 to not recycle.

    @ivar ampChild: The child AMP protocol subclass with the commands
                    that the child should implement.

    @ivar ampParent: The parent AMP protocol subclass with the commands
                    that the parent should implement.

    @ivar timeout: The general timeout (in seconds) for every child
                    process call.
    """

    finished = False
    started = False
    name = None

    def __init__(self, ampChild=None, ampParent=None, min=5, max=20,
                 name=None, maxIdle=20, recycleAfter=500, starter=None,
                 timeout=None, timeout_signal=DIE, ampChildArgs=()):
        self.starter = starter
        self.ampChildArgs = tuple(ampChildArgs)
        if starter is None:
            self.starter = main.ProcessStarter(packages=("twisted",))
        self.ampParent = ampParent
        self.ampChild = ampChild
        if ampChild is None:
            from ampoule.child import AMPChild
            self.ampChild = AMPChild
        self.min = min
        self.max = max
        self.name = name
        self.maxIdle = maxIdle
        self.recycleAfter = recycleAfter
        self.timeout = timeout
        self.timeout_signal = timeout_signal
        self._queue = []

        self.processes = set()
        self.ready = set()
        self.busy = set()
        self._finishCallbacks = {}
        self._lastUsage = {}
        self._calls = {}
        self.looping = task.LoopingCall(self._pruneProcesses)
        self.looping.start(maxIdle, now=False)

    def start(self, ampChild=None):
        """
        Starts the ProcessPool with a given child protocol.

        @param ampChild: a L{ampoule.child.AMPChild} subclass.
        @type ampChild: L{ampoule.child.AMPChild} subclass
        """
        if ampChild is not None and not self.started:
            self.ampChild = ampChild
        self.finished = False
        self.started = True
        return self.adjustPoolSize()

    def _pruneProcesses(self):
        """
        Remove idle processes from the pool.
        """
        n = now()
        d = []
        for child, lastUse in self._lastUsage.items():
            if len(self.processes) > self.min and (n - lastUse) > self.maxIdle:
                # we are setting lastUse when processing finishes, it
                # might be processing right now
                if child not in self.busy:
                    # we need to remove this child from the ready set
                    # and the processes set because otherwise it might
                    # get calls from doWork
                    self.ready.discard(child)
                    self.processes.discard(child)
                    d.append(self.stopAWorker(child))
        return defer.DeferredList(d)

    def _pruneProcess(self, child):
        """
        Remove every trace of the process from this instance.
        """
        self.processes.discard(child)
        self.ready.discard(child)
        self.busy.discard(child)
        self._lastUsage.pop(child, None)
        self._calls.pop(child, None)
        self._finishCallbacks.pop(child, None)

    def _addProcess(self, child, finished):
        """
        Adds the newly created child process to the pool.
        """
        def fatal(reason, child):
            log.error(
                u'FATAL: Process exited.\n\t{r}', r=reason.getErrorMessage()
            )
            self._pruneProcess(child)

        def dieGently(data, child):
            log.info(u'STOPPING: {s}', s=data)
            self._pruneProcess(child)

        self.processes.add(child)
        self.ready.add(child)
        finished.addCallback(dieGently, child).addErrback(fatal, child)
        self._finishCallbacks[child] = finished
        self._lastUsage[child] = now()
        self._calls[child] = 0
        self._catchUp()

    def _catchUp(self):
        """
        If there are queued items in the list then run them.
        """
        if self._queue:
            _, (d, command, kwargs) = pop(self._queue)
            self._cb_doWork(command, **kwargs).chainDeferred(d)

    def _handleTimeout(self, child):
        """
        One of the children went timeout, we need to deal with it

        @param child: The child process
        @type child: L{child.AMPChild}
        """
        try:
            child.transport.signalProcess(self.timeout_signal)
        except error.ProcessExitedAlready:
            # don't do anything then... we are too late
            # or we were too early to call
            pass

    def startAWorker(self):
        """
        Start a worker and set it up in the system.
        """
        if self.finished:
            # this is a race condition: basically if we call self.stop()
            # while a process is being recycled what happens is that the
            # process will be created anyway. By putting a check for
            # self.finished here we make sure that in no way we are creating
            # processes when the pool is stopped.
            # The race condition comes from the fact that:
            # stopAWorker() is asynchronous while stop() is synchronous.
            # so if you call:
            # pp.stopAWorker(child).addCallback(lambda _: pp.startAWorker())
            # pp.stop()
            # You might end up with a dirty reactor due to the stop()
            # returning before the new process is created.
            return
        startAMPProcess = self.starter.startAMPProcess
        child, finished = startAMPProcess(self.ampChild,
                                          ampParent=self.ampParent,
                                          ampChildArgs=self.ampChildArgs)
        return self._addProcess(child, finished)

    def _cb_doWork(self, command, _timeout=None, _deadline=None,
                   **kwargs):
        """
        Go and call the command.

        @param command: The L{amp.Command} to be executed in the child
        @type command: L{amp.Command}

        @param _d: The deferred for the calling code.
        @type _d: L{defer.Deferred}

        @param _timeout: The timeout for this call only
        @type _timeout: C{int}
        @param _deadline: The deadline for this call only
        @type _deadline: C{int}
        """
        timeoutCall = None
        deadlineCall = None

        def _returned(result, child, is_error=False):
            def cancelCall(call):
                if call is not None and call.active():
                    call.cancel()
            cancelCall(timeoutCall)
            cancelCall(deadlineCall)
            self.busy.discard(child)

            alreadyDead = (
                is_error and
                result.check(error.ProcessTerminated)
            )

            if not (die or alreadyDead):
                # we are not marked to be removed, so add us back to
                # the ready set and let's see if there's some catching
                # up to do
                self.ready.add(child)
                self._catchUp()
            else:
                # We should die and we do, then we start a new worker
                # to pick up stuff from the queue otherwise we end up
                # without workers and the queue will remain there.
                self.stopAWorker(child).addCallback(lambda _: self.startAWorker())
            self._lastUsage[child] = now()
            # we can't do recycling here because it's too late and
            # the process might have received tons of calls already
            # which would make it run more calls than what is
            # configured to do.
            return result

        die = False
        child = self.ready.pop()
        self.busy.add(child)
        self._calls[child] += 1

        # Let's see if this call goes over the recycling barrier
        if self.recycleAfter and self._calls[child] >= self.recycleAfter:
            # it does so mark this child, using a closure, to be
            # removed at the end of the call.
            die = True

        # If the command doesn't require a response then callRemote
        # returns nothing, so we prepare for that too.
        # We also need to guard against timeout errors for child
        # and local timeout parameter overrides the global one
        if _timeout == 0:
            timeout = _timeout
        else:
            timeout = _timeout or self.timeout

        if timeout is not None:
            from twisted.internet import reactor
            timeoutCall = reactor.callLater(timeout, self._handleTimeout, child)

        if _deadline is not None:
            from twisted.internet import reactor
            delay = max(0, _deadline - reactor.seconds())
            deadlineCall = reactor.callLater(delay, self._handleTimeout,
                                             child)

        return defer.maybeDeferred(child.callRemote, command, **kwargs
            ).addCallback(_returned, child
            ).addErrback(_returned, child, is_error=True)

    def callRemote(self, *args, **kwargs):
        """
        Proxy call to keep the API homogeneous across twisted's RPCs
        """
        return self.doWork(*args, **kwargs)

    def doWork(self, command, **kwargs):
        """
        Sends the command to one child.

        @param command: an L{amp.Command} type object.
        @type command: L{amp.Command}

        @param kwargs: dictionary containing the arguments for the command.
        """
        if self.ready: # there are unused processes, let's use them
            return self._cb_doWork(command, **kwargs)
        else:
            if len(self.processes) < self.max:
                # no unused but we can start some new ones
                # since startAWorker is synchronous we won't have a
                # race condition here in case of multiple calls to
                # doWork, so we will end up in the else clause in case
                # of such calls:
                # Process pool with min=1, max=1, recycle_after=1
                # [call(Command) for x in xrange(BIG_NUMBER)]
                self.startAWorker()
                return self._cb_doWork(command, **kwargs)
            else:
                # No one is free... just queue up and wait for a process
                # to start and pick up the first item in the queue.
                d = defer.Deferred()
                self._queue.append((count(), (d, command, kwargs)))
                return d

    def stopAWorker(self, child=None):
        """
        Gently stop a child so that it's not restarted anymore

        @param child: an L{ampoule.child.AmpChild} type object.
        @type child: L{ampoule.child.AmpChild} or None

        """
        if child is None:
            if self.ready:
                child = self.ready.pop()
            else:
                child = choice(list(self.processes))
        child.callRemote(commands.Shutdown
            # This is needed for timeout handling, the reason is pretty hard
            # to explain but I'll try to:
            # There's another small race condition in the system. If the
            # child process is shut down by a signal and you try to stop
            # the process pool immediately afterwards, like tests would do,
            # the child AMP object would still be in the system and trying
            # to call the command Shutdown on it would result in the same
            # errback that we got originally, for this reason we need to
            # trap it now so that it doesn't raise by not being handled.
            # Does this even make sense to you?
            ).addErrback(lambda reason: reason.trap(error.ProcessTerminated))
        return self._finishCallbacks[child]

    def adjustPoolSize(self, min=None, max=None):
        """
        Change the pool size to be at least min and less than max,
        useful when you change the values of max and min in the instance
        and you want the pool to adapt to them.
        """
        if min is None:
            min = self.min
        if max is None:
            max = self.max

        assert min >= 0, 'minimum is negative'
        assert min <= max, 'minimum is greater than maximum'

        self.min = min
        self.max = max

        l = []
        if self.started:

            for i in range(len(self.processes)-self.max):
                l.append(self.stopAWorker())
            while len(self.processes) < self.min:
                self.startAWorker()

        return defer.DeferredList(l).addCallback(lambda _: self.dumpStats())

    def stop(self):
        """
        Stops the process protocol.
        """
        self.finished = True
        l = [self.stopAWorker(process) for process in self.processes]
        def _cb(_):
            if self.looping.running:
                self.looping.stop()

        return defer.DeferredList(l).addCallback(_cb)

    def dumpStats(self):
        log.info(
            STATS_TEMPLATE,
            w=len(self.processes),
            t=self.timeout,
            p=self.ampParent,
            c=self.ampChild,
            i=self.maxIdle,
            r=self.recycleAfter,
            s=self.starter
        )

pp = None

def deferToAMPProcess(command, **kwargs):
    """
    Helper function that sends a command to the default process pool
    and returns a deferred that fires when the result of the
    subprocess computation is ready.

    @param command: an L{amp.Command} subclass
    @param kwargs: dictionary containing the arguments for the command.

    @return: a L{defer.Deferred} with the data from the subprocess.
    """
    global pp
    if pp is None:
        pp = ProcessPool()
        return pp.start().addCallback(lambda _: pp.doWork(command, **kwargs))
    return pp.doWork(command, **kwargs)




from signal import SIGHUP
import os
import os.path
from io import BytesIO as sio
import tempfile

from twisted.internet import error, defer, reactor
from twisted.python import failure
from twisted.trial import unittest
from twisted.protocols import amp
from ampoule import main, child, commands, pool

class ShouldntHaveBeenCalled(Exception):
    pass

def _raise(_):
    raise ShouldntHaveBeenCalled(_)

class _FakeT(object):
    closeStdinCalled = False
    def __init__(self, s):
        self.s = s

    def closeStdin(self):
        self.closeStdinCalled = True

    def write(self, data):
        self.s.write(data)

class FakeAMP(object):
    connector = None
    reason = None
    def __init__(self, s):
        self.s = s

    def makeConnection(self, connector):
        if self.connector is not None:
            raise Exception("makeConnection called twice")
        self.connector = connector

    def connectionLost(self, reason):
        if self.reason is not None:
            raise Exception("connectionLost called twice")
        self.reason = reason

    def dataReceived(self, data):
        self.s.write(data)

class Exit(amp.Command):
    arguments = []
    response = []

class Ping(amp.Command):
    arguments = [(b'data', amp.String())]
    response = [(b'response', amp.String())]

class Pong(amp.Command):
    arguments = [(b'data', amp.String())]
    response = [(b'response', amp.String())]

class Pid(amp.Command):
    response = [(b'pid', amp.Integer())]

class Reactor(amp.Command):
    response = [(b'classname', amp.String())]

class NoResponse(amp.Command):
    arguments = [(b'arg', amp.String())]
    requiresAnswer = False

class GetResponse(amp.Command):
    response = [(b"response", amp.String())]

class Child(child.AMPChild):
    @Ping.responder
    def ping(self, data):
        return self.callRemote(Pong, data=data)

class PidChild(child.AMPChild):
    @Pid.responder
    def pid(self):
        import os
        return {'pid': os.getpid()}

class NoResponseChild(child.AMPChild):
    _set = False
    @NoResponse.responder
    def noresponse(self, arg):
        self._set = arg
        return {}

    @GetResponse.responder
    def getresponse(self):
        return {"response": self._set}

class ReactorChild(child.AMPChild):
    @Reactor.responder
    def reactor(self):
        from twisted.internet import reactor
        return {'classname': reactor.__class__.__name__.encode()}

class First(amp.Command):
    arguments = [(b'data', amp.String())]
    response = [(b'response', amp.String())]

class Second(amp.Command):
    pass

class WaitingChild(child.AMPChild):
    deferred = None
    @First.responder
    def first(self, data):
        self.deferred = defer.Deferred()
        return self.deferred.addCallback(lambda _: {'response': data})

    @Second.responder
    def second(self):
        self.deferred.callback('')
        return {}

class HangForever(amp.Command):
    pass

class TimingOutChild(child.AMPChild):
    @HangForever.responder
    def hangForever(self):
        return defer.Deferred()

    @Ping.responder
    def ping(self, data):
        return {'response': data}

class Die(amp.Command):
    pass

class BadChild(child.AMPChild):
    @Die.responder
    def die(self):
        self.shutdown = False
        self.transport.loseConnection()
        return {}


class ExitingChild(child.AMPChild):
    @Exit.responder
    def exit(self):
        import os
        os._exit(33)

class Write(amp.Command):
    response = [(b"response", amp.String())]


class Writer(child.AMPChild):

    def __init__(self, data=b'hello'):
        child.AMPChild.__init__(self)
        if isinstance(data, str):
            # this is passing through sys.argv, argv is unconditionally unicode
            # on py3; see https://bugs.python.org/issue8776
            data = data.encode()
        self.data = data

    @Write.responder
    def write(self):
        return {'response': self.data}


class GetCWD(amp.Command):

    response = [(b"cwd", amp.Unicode())]


class TempDirChild(child.AMPChild):

    def __init__(self, directory=None):
        child.AMPChild.__init__(self)
        self.directory = directory

    def __enter__(self):
        directory = tempfile.mkdtemp()
        os.chdir(directory)
        if self.directory is not None:
            os.mkdir(self.directory)
            os.chdir(self.directory)

    def __exit__(self, exc_type, exc_val, exc_tb):
        cwd = os.getcwd()
        os.chdir('..')
        os.rmdir(cwd)

    @GetCWD.responder
    def getcwd(self):
        return {'cwd': os.getcwd()}


class TestAMPConnector(unittest.TestCase):
    def setUp(self):
        """
        The only reason why this method exists is to let 'trial ampoule'
        to install the signal handlers (#3178 for reference).
        """
        super(TestAMPConnector, self).setUp()
        d = defer.Deferred()
        reactor.callLater(0, d.callback, None)
        return d

    def _makeConnector(self, s, sa):
        a = FakeAMP(sa)
        ac = main.AMPConnector(a)
        assert ac.name is not None
        ac.transport = _FakeT(s)
        return ac

    def test_protocol(self):
        """
        Test that outReceived writes to AMP and that it triggers the
        finished deferred once the process ended.
        """
        s = sio()
        sa = sio()
        ac = self._makeConnector(s, sa)

        for x in range(99):
            ac.childDataReceived(4, str(x).encode("ascii"))

        ac.processEnded(failure.Failure(error.ProcessDone(0)))
        return ac.finished.addCallback(
            lambda _: self.assertEqual(sa.getvalue(), b''.join(
                str(x).encode("ascii") for x in range(99)
            ))
        )

    def test_protocol_failing(self):
        """
        Test that a failure in the process termination is correctly
        propagated to the finished deferred.
        """
        s = sio()
        sa = sio()
        ac = self._makeConnector(s, sa)

        ac.finished.addCallback(_raise)
        fail = failure.Failure(error.ProcessTerminated())
        self.assertFailure(ac.finished, error.ProcessTerminated)
        ac.processEnded(fail)

    def test_startProcess(self):
        """
        Test that startProcess actually starts a subprocess and that
        it receives data back from the process through AMP.
        """
        s = sio()
        a = FakeAMP(s)
        STRING = b"ciao"
        BOOT = """\
import sys, os
def main(arg):
    os.write(4, arg.encode("utf-8"))
main(sys.argv[1])
"""
        starter = main.ProcessStarter(bootstrap=BOOT,
                                      args=(STRING,),
                                      packages=("twisted", "ampoule"))

        amp, finished = starter.startPythonProcess(main.AMPConnector(a))
        return finished.addCallback(lambda _: self.assertEquals(s.getvalue(), STRING))

    def test_failing_deferToProcess(self):
        """
        Test failing subprocesses and the way they terminate and preserve
        failing information.
        """
        s = sio()
        a = FakeAMP(s)
        STRING = b"ciao"
        BOOT = """\
import sys
def main(arg):
    raise Exception(arg)
main(sys.argv[1])
"""
        starter = main.ProcessStarter(bootstrap=BOOT, args=(STRING,), packages=("twisted", "ampoule"))
        ready, finished = starter.startPythonProcess(main.AMPConnector(a), "I'll be ignored")

        self.assertFailure(finished, error.ProcessTerminated)
        finished.addErrback(lambda reason: self.assertEquals(reason.getMessage(), STRING))
        return finished

    def test_env_setting(self):
        """
        Test that and environment variable passed to the process starter
        is correctly passed to the child process.
        """
        s = sio()
        a = FakeAMP(s)
        STRING = b"ciao"
        BOOT = """\
import sys, io, os
def main():
    with io.open(4, 'w' + ('b' if bytes is str else '')) as f:
        f.write(os.environ['FOOBAR'])
main()
"""
        starter = main.ProcessStarter(bootstrap=BOOT,
                                      packages=("twisted", "ampoule"),
                                      env={"FOOBAR": STRING})
        amp, finished = starter.startPythonProcess(main.AMPConnector(a), "I'll be ignored")
        return finished.addCallback(lambda _: self.assertEquals(s.getvalue(), STRING))

    def test_startAMPProcess(self):
        """
        Test that you can start an AMP subprocess and that it correctly
        accepts commands and correctly answers them.
        """
        STRING = b"ciao"

        starter = main.ProcessStarter(packages=("twisted", "ampoule"))
        c, finished = starter.startAMPProcess(child.AMPChild)
        c.callRemote(commands.Echo, data=STRING
           ).addCallback(lambda response:
                self.assertEquals(response['response'], STRING)
           ).addCallback(lambda _: c.callRemote(commands.Shutdown))
        return finished

    def test_BootstrapContext(self):
        starter = main.ProcessStarter(packages=('twisted', 'ampoule'))
        c, finished = starter.startAMPProcess(TempDirChild)
        cwd = []
        def checkBootstrap(response):
            cwd.append(response['cwd'])
            self.assertNotEquals(cwd, os.getcwd())
        d = c.callRemote(GetCWD)
        d.addCallback(checkBootstrap)
        d.addCallback(lambda _: c.callRemote(commands.Shutdown))
        finished.addCallback(lambda _: self.assertFalse(os.path.exists(cwd[0])))
        return finished

    def test_BootstrapContextInstance(self):
        starter = main.ProcessStarter(packages=('twisted', 'ampoule'))
        c, finished = starter.startAMPProcess(TempDirChild,
                                              ampChildArgs=('foo',))
        cwd = []
        def checkBootstrap(response):
            cwd.append(response['cwd'])
            self.assertTrue(cwd[0].endswith('/foo'))
        d = c.callRemote(GetCWD)
        d.addCallback(checkBootstrap)
        d.addCallback(lambda _: c.callRemote(commands.Shutdown))
        finished.addCallback(lambda _: self.assertFalse(os.path.exists(cwd[0])))
        return finished

    def test_startAMPAndParentProtocol(self):
        """
        Test that you can start an AMP subprocess and the children can
        call methods on their parent.
        """
        DATA = b"CIAO"
        APPEND = b"123"

        class Parent(amp.AMP):
            @Pong.responder
            def pong(self, data):
                return {'response': DATA+APPEND}

        starter = main.ProcessStarter(packages=("twisted", "ampoule"))

        subp, finished = starter.startAMPProcess(ampChild=Child, ampParent=Parent)
        subp.callRemote(Ping, data=DATA
           ).addCallback(lambda response:
                self.assertEquals(response['response'], DATA+APPEND)
           ).addCallback(lambda _: subp.callRemote(commands.Shutdown))
        return finished

    def test_roundtripError(self):
        """
        Test that invoking a child using an unreachable class raises
        a L{RunTimeError} .
        """
        class Child(child.AMPChild):
            pass

        starter = main.ProcessStarter(packages=("twisted", "ampoule"))

        self.assertRaises(RuntimeError, starter.startAMPProcess, ampChild=Child)

class TestProcessPool(unittest.TestCase):

    def test_startStopWorker(self):
        """
        Test that starting and stopping a worker keeps the state of
        the process pool consistent.
        """
        pp = pool.ProcessPool()
        self.assertEquals(pp.started, False)
        self.assertEquals(pp.finished, False)
        self.assertEquals(pp.processes, set())
        self.assertEquals(pp._finishCallbacks, {})

        def _checks():
            self.assertEquals(pp.started, False)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), 1)
            self.assertEquals(len(pp._finishCallbacks), 1)
            return pp.stopAWorker()

        def _closingUp(_):
            self.assertEquals(pp.started, False)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), 0)
            self.assertEquals(pp._finishCallbacks, {})
        pp.startAWorker()
        return _checks().addCallback(_closingUp).addCallback(lambda _: pp.stop())

    def test_startAndStop(self):
        """
        Test that a process pool's start and stop method create the
        expected number of workers and keep state consistent in the
        process pool.
        """
        pp = pool.ProcessPool()
        self.assertEquals(pp.started, False)
        self.assertEquals(pp.finished, False)
        self.assertEquals(pp.processes, set())
        self.assertEquals(pp._finishCallbacks, {})

        def _checks(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)
            return pp.stop()

        def _closingUp(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, True)
            self.assertEquals(len(pp.processes), 0)
            self.assertEquals(pp._finishCallbacks, {})
        return pp.start().addCallback(_checks).addCallback(_closingUp)

    def test_adjustPoolSize(self):
        """
        Test that calls to pool.adjustPoolSize are correctly handled.
        """
        pp = pool.ProcessPool(min=10)
        self.assertEquals(pp.started, False)
        self.assertEquals(pp.finished, False)
        self.assertEquals(pp.processes, set())
        self.assertEquals(pp._finishCallbacks, {})

        def _resize1(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)
            return pp.adjustPoolSize(min=2, max=3)

        def _resize2(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(pp.max, 3)
            self.assertEquals(pp.min, 2)
            self.assertEquals(len(pp.processes), pp.max)
            self.assertEquals(len(pp._finishCallbacks), pp.max)

        def _resize3(_):
            self.assertRaises(AssertionError, pp.adjustPoolSize, min=-1, max=5)
            self.assertRaises(AssertionError, pp.adjustPoolSize, min=5, max=1)
            return pp.stop()

        return pp.start(
            ).addCallback(_resize1
            ).addCallback(_resize2
            ).addCallback(_resize3)

    def test_childRestart(self):
        """
        Test that a failing child process is immediately restarted.
        """
        pp = pool.ProcessPool(ampChild=BadChild, min=1)
        STRING = b"DATA"

        def _checks(_):
            d = next(iter(pp._finishCallbacks.values()))
            pp.doWork(Die).addErrback(lambda _: None)
            return d.addBoth(_checksAgain)

        def _checksAgain(_):
            return pp.doWork(commands.Echo, data=STRING
                    ).addCallback(lambda result: self.assertEquals(result['response'], STRING))

        return pp.start(
            ).addCallback(_checks
            ).addCallback(lambda _: pp.stop())

    def test_parentProtocolChange(self):
        """
        Test that the father can use an AMP protocol too.
        """
        DATA = b"CIAO"
        APPEND = b"123"

        class Parent(amp.AMP):
            @Pong.responder
            def pong(self, data):
                return {'response': DATA+APPEND}

        pp = pool.ProcessPool(ampChild=Child, ampParent=Parent)
        def _checks(_):
            return pp.doWork(Ping, data=DATA
                       ).addCallback(lambda response:
                            self.assertEquals(response['response'], DATA+APPEND)
                       )

        return pp.start().addCallback(_checks).addCallback(lambda _: pp.stop())


    def test_deferToAMPProcess(self):
        """
        Test that deferToAMPProcess works as expected.
        """
        def cleanupGlobalPool():
            d = pool.pp.stop()
            pool.pp = None
            return d
        self.addCleanup(cleanupGlobalPool)

        STRING = b"CIAOOOO"
        d = pool.deferToAMPProcess(commands.Echo, data=STRING)
        d.addCallback(self.assertEquals, {"response": STRING})
        return d

    def test_checkStateInPool(self):
        """
        Test that busy and ready lists are correctly maintained.
        """
        pp = pool.ProcessPool(ampChild=WaitingChild)

        DATA = b"foobar"

        def _checks(_):
            d = pp.callRemote(First, data=DATA)
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)
            self.assertEquals(len(pp.ready), pp.min-1)
            self.assertEquals(len(pp.busy), 1)
            child = pp.busy.pop()
            pp.busy.add(child)
            child.callRemote(Second)
            return d

        return pp.start(
            ).addCallback(_checks
            ).addCallback(lambda _: pp.stop())

    def test_growingToMax(self):
        """
        Test that the pool grows over time until it reaches max processes.
        """
        MAX = 5
        pp = pool.ProcessPool(ampChild=WaitingChild, min=1, max=MAX)

        def _checks(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)

            D = b"DATA"
            d = [pp.doWork(First, data=D) for x in range(MAX)]

            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.max)
            self.assertEquals(len(pp._finishCallbacks), pp.max)

            [child.callRemote(Second) for child in pp.processes]
            return defer.DeferredList(d)

        return pp.start(
            ).addCallback(_checks
            ).addCallback(lambda _: pp.stop())

    def test_growingToMaxAndShrinking(self):
        """
        Test that the pool grows but after 'idle' time the number of
        processes goes back to the minimum.
        """

        MAX = 5
        MIN = 1
        IDLE = 1
        pp = pool.ProcessPool(ampChild=WaitingChild, min=MIN, max=MAX, maxIdle=IDLE)

        def _checks(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)

            D = b"DATA"
            d = [pp.doWork(First, data=D) for x in range(MAX)]

            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.max)
            self.assertEquals(len(pp._finishCallbacks), pp.max)

            [child.callRemote(Second) for child in pp.processes]
            return defer.DeferredList(d).addCallback(_realChecks)

        def _realChecks(_):
            from twisted.internet import reactor
            d = defer.Deferred()
            def _cb():
                def __(_):
                    try:
                        self.assertEquals(pp.started, True)
                        self.assertEquals(pp.finished, False)
                        self.assertEquals(len(pp.processes), pp.min)
                        self.assertEquals(len(pp._finishCallbacks), pp.min)
                        d.callback(None)
                    except Exception as e:
                        d.errback(e)
                return pp._pruneProcesses().addCallback(__)
            # just to be shure we are called after the pruner
            pp.looping.stop() # stop the looping, we don't want it to
                              # this right here
            reactor.callLater(IDLE, _cb)
            return d

        return pp.start(
            ).addCallback(_checks
            ).addCallback(lambda _: pp.stop())

    def test_recycling(self):
        """
        Test that after a given number of calls subprocesses are
        recycled.
        """
        MAX = 1
        MIN = 1
        RECYCLE_AFTER = 1
        pp = pool.ProcessPool(ampChild=PidChild, min=MIN, max=MAX, recycleAfter=RECYCLE_AFTER)
        self.addCleanup(pp.stop)

        def _checks(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)
            return pp.doWork(Pid
                ).addCallback(lambda response: response['pid'])

        def _checks2(pid):
            return pp.doWork(Pid
                ).addCallback(lambda response: response['pid']
                ).addCallback(self.assertNotEquals, pid)


        d = pp.start()
        d.addCallback(_checks)
        d.addCallback(_checks2)
        return d

    def test_recyclingProcessFails(self):
        """
        A process exiting with a non-zero exit code when recycled does not get
        multiple processes started to replace it.
        """
        MAX = 1
        MIN = 1
        RECYCLE_AFTER = 1
        RECYCLE_AFTER = 1
        pp = pool.ProcessPool(ampChild=ExitingChild, min=MIN, max=MAX, recycleAfter=RECYCLE_AFTER)
        self.addCleanup(pp.stop)

        def _checks(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)
            child = list(pp.ready)[0]
            finished = pp._finishCallbacks[child]
            return pp.doWork(Exit).addBoth(lambda _: finished)

        def _checks2(_):
            self.assertEquals(len(pp.processes), pp.max)

        d = pp.start()
        d.addCallback(_checks)
        d.addCallback(_checks2)
        return d


    def test_recyclingWithQueueOverload(self):
        """
        Test that we get the correct number of different results when
        we overload the pool of calls.
        """
        MAX = 5
        MIN = 1
        RECYCLE_AFTER = 10
        CALLS = 60
        pp = pool.ProcessPool(ampChild=PidChild, min=MIN, max=MAX, recycleAfter=RECYCLE_AFTER)
        self.addCleanup(pp.stop)

        def _check(results):
            s = set()
            for succeed, response in results:
                s.add(response['pid'])

            # For the first C{MAX} calls, each is basically guaranteed to go to
            # a different child.  After that, though, there are no guarantees.
            # All the rest might go to a single child, since the child to
            # perform a job is selected arbitrarily from the "ready" set.  Fair
            # distribution of jobs needs to be implemented; right now it's "set
            # ordering" distribution of jobs.
            self.assertTrue(len(s) > MAX)

        def _work(_):
            l = [pp.doWork(Pid) for x in range(CALLS)]
            d = defer.DeferredList(l)
            return d.addCallback(_check)
        d = pp.start()
        d.addCallback(_work)
        return d


    def test_disableProcessRecycling(self):
        """
        Test that by setting 0 to recycleAfter we actually disable process recycling.
        """
        MAX = 1
        MIN = 1
        RECYCLE_AFTER = 0
        pp = pool.ProcessPool(ampChild=PidChild, min=MIN, max=MAX, recycleAfter=RECYCLE_AFTER)

        def _checks(_):
            self.assertEquals(pp.started, True)
            self.assertEquals(pp.finished, False)
            self.assertEquals(len(pp.processes), pp.min)
            self.assertEquals(len(pp._finishCallbacks), pp.min)
            return pp.doWork(Pid
                ).addCallback(lambda response: response['pid'])

        def _checks2(pid):
            return pp.doWork(Pid
                ).addCallback(lambda response: response['pid']
                ).addCallback(self.assertEquals, pid
                ).addCallback(lambda _: pid)

        def finish(reason):
            return pp.stop().addCallback(lambda _: reason)

        return pp.start(
            ).addCallback(_checks
            ).addCallback(_checks2
            ).addCallback(_checks2
            ).addCallback(finish)

    def test_changeChildrenReactor(self):
        """
        Test that by passing the correct argument children change their
        reactor type.
        """
        MAX = 1
        MIN = 1
        FIRST = "select"
        SECOND = "poll"

        def checkDefault():
            pp = pool.ProcessPool(
                starter=main.ProcessStarter(
                    childReactor=FIRST,
                    packages=("twisted", "ampoule")),
                ampChild=ReactorChild, min=MIN, max=MAX)
            pp.start()
            return (pp.doWork(Reactor)
                    .addCallback(self.assertEquals,
                                 {'classname': b"SelectReactor"})
                    .addCallback(lambda _: pp.stop()))
        def checkPool(_):
            pp = pool.ProcessPool(
                starter=main.ProcessStarter(
                    childReactor=SECOND,
                    packages=("twisted", "ampoule")),
                ampChild=ReactorChild, min=MIN, max=MAX)
            pp.start()
            return (pp.doWork(Reactor)
                    .addCallback(self.assertEquals,
                                 {'classname': b"PollReactor"})
                    .addCallback(lambda _: pp.stop()))

        return checkDefault(
            ).addCallback(checkPool)
    try:
        from select import poll
    except ImportError:
        test_changeChildrenReactor.skip = "This architecture doesn't support select.poll, I can't run this test"

    def test_commandsWithoutResponse(self):
        """
        Test that if we send a command without a required answer we
        actually don't have any problems.
        """
        DATA = b"hello"
        pp = pool.ProcessPool(ampChild=NoResponseChild, min=1, max=1)

        def _check(_):
            return pp.doWork(GetResponse
                ).addCallback(self.assertEquals, {"response": DATA})

        def _work(_):
            return pp.doWork(NoResponse, arg=DATA)

        return pp.start(
            ).addCallback(_work
            ).addCallback(_check
            ).addCallback(lambda _: pp.stop())

    def test_supplyChildArgs(self):
        """Ensure that arguments for the child constructor are passed in."""
        pp = pool.ProcessPool(Writer, ampChildArgs=['body'], min=0)
        def _check(result):
            return pp.doWork(Write).addCallback(
            self.assertEquals, {'response': b'body'})

        return pp.start(
            ).addCallback(_check
            ).addCallback(lambda _: pp.stop())

    def processTimeoutTest(self, timeout):
        pp = pool.ProcessPool(WaitingChild, min=1, max=1)

        def _work(_):
            d = pp.callRemote(First, data=b"ciao", _timeout=timeout)
            self.assertFailure(d, error.ProcessTerminated)
            return d

        return pp.start(
            ).addCallback(_work
            ).addCallback(lambda _: pp.stop())

    def test_processTimeout(self):
        """
        Test that a call that doesn't finish within the given timeout
        time is correctly handled.
        """
        return self.processTimeoutTest(1)

    def test_processTimeoutZero(self):
        """
        Test that the process is correctly handled when the timeout is zero.
        """
        return self.processTimeoutTest(0)

    def test_processDeadline(self):
        pp = pool.ProcessPool(WaitingChild, min=1, max=1)

        def _work(_):
            d = pp.callRemote(First, data=b"ciao", _deadline=reactor.seconds())
            self.assertFailure(d, error.ProcessTerminated)
            return d

        return pp.start(
            ).addCallback(_work
            ).addCallback(lambda _: pp.stop())

    def test_processBeforeDeadline(self):
        pp = pool.ProcessPool(PidChild, min=1, max=1)

        def _work(_):
            d = pp.callRemote(Pid, _deadline=reactor.seconds() + 10)
            d.addCallback(lambda result: self.assertNotEqual(result['pid'], 0))
            return d

        return pp.start(
            ).addCallback(_work
            ).addCallback(lambda _: pp.stop())

    def test_processTimeoutSignal(self):
        """
        Test that a call that doesn't finish within the given timeout
        time is correctly handled.
        """
        pp = pool.ProcessPool(WaitingChild, min=1, max=1,
                              timeout_signal=SIGHUP)

        def _work(_):
            d = pp.callRemote(First, data=b"ciao", _timeout=1)
            d.addCallback(lambda d: self.fail())
            text = 'signal %d' % SIGHUP
            d.addErrback(lambda f: self.assertIn(text, str(f.value)))
            return d

        return pp.start(
            ).addCallback(_work
            ).addCallback(lambda _: pp.stop())

    def test_processGlobalTimeout(self):
        """
        Test that a call that doesn't finish within the given global
        timeout time is correctly handled.
        """
        pp = pool.ProcessPool(WaitingChild, min=1, max=1, timeout=1)

        def _work(_):
            d = pp.callRemote(First, data=b"ciao")
            self.assertFailure(d, error.ProcessTerminated)
            return d

        return pp.start(
            ).addCallback(_work
            ).addCallback(lambda _: pp.stop())

    def test_processRestartAfterTimeout(self):
        """
        Test that a call that times out doesn't cause all subsequent requests
        to fail
        """
        pp = pool.ProcessPool(TimingOutChild, min=1, max=1, timeout=1)

        def _work(_):
            d1 = pp.callRemote(HangForever)
            d2 = pp.callRemote(Ping, data=b"hello")

            self.assertFailure(d1, error.ProcessTerminated)

            d2.addCallback(
                lambda result: self.assertEqual(
                    result, {"response": b"hello"}))

            return defer.DeferredList([d1, d2])

        return pp.start(
            ).addCallback(_work
            ).addCallback(lambda _: pp.stop())



from ampoule import child, util
from twisted.protocols import amp

class Pid(amp.Command):
    response = [("pid", amp.Integer())]

class MyChild(child.AMPChild):
    @Pid.responder
    def pid(self):
        import os
        return {"pid": os.getpid()}

@util.mainpoint
def main(args):
    import sys
    from twisted.internet import reactor, defer
    from twisted.python import log
    log.startLogging(sys.stdout)

    from ampoule import pool

    @defer.inlineCallbacks
    def _run():
        pp = pool.ProcessPool(MyChild, min=1, max=1)
        yield pp.start()
        result = yield pp.doWork(Pid)
        print("The Child process PID is:", result['pid'])
        yield pp.stop()
        reactor.stop()

    reactor.callLater(1, _run)
    reactor.run()



from ampoule import child, util

@util.mainpoint
def main(args):
    from twisted.internet import reactor, defer
    from ampoule import pool, Ping
    import time

    # enable logging to see what happens in children
    import sys
    from twisted.python import log
    log.startLogging(sys.stdout)

    @defer.inlineCallbacks
    def _run():
        pp = pool.ProcessPool(child.AMPChild, recycleAfter=5000)
        pp.min = 1
        pp.max = 5
        yield pp.start()
        t = time.time()
        REPEATS = 40000
        l = [pp.doWork(Ping) for x in range(REPEATS)]
        yield defer.DeferredList(l)
        print(REPEATS/(time.time() - t))
        yield pp.stop()
        reactor.stop()

    reactor.callLater(1, _run)
    reactor.run()



from twisted import logger
from twisted.internet import error
from twisted.protocols import amp
from ampoule.commands import Echo, Shutdown, Ping



log = logger.Logger()


class AMPChild(amp.AMP):
    def __init__(self):
        super(AMPChild, self).__init__(self)
        self.shutdown = False

    def connectionLost(self, reason):
        amp.AMP.connectionLost(self, reason)
        from twisted.internet import reactor
        try:
            reactor.stop()
        except error.ReactorNotRunning:
            # woa, this means that something bad happened,
            # most probably we received a SIGINT. Now this is only
            # a problem when you use Ctrl+C to stop the main process
            # because it would send the SIGINT to child processes too.
            # In all other cases receiving a SIGINT here would be an
            # error condition and correctly restarted. maybe we should
            # use sigprocmask?
            pass
        if not self.shutdown:
            # if the shutdown wasn't explicit we presume that it's an
            # error condition and thus we return a -1 error returncode.
            import os
            os._exit(-1)

    def shutdown(self):
        """
        This method is needed to shutdown the child gently without
        generating an exception.
        """
        log.info(u'Shutdown message received, goodbye.')
        self.shutdown = True
        return {}
    Shutdown.responder(shutdown)

    def ping(self):
        """
        Ping the child and return an answer
        """
        return {'response': "pong"}
    Ping.responder(ping)

    def echo(self, data):
        """
        Echo some data through the child.
        """
        return {'response': data}
    Echo.responder(echo)



"""
This module implements a remote pool to use with AMP.
"""
from twisted.protocols import amp

class AMPProxy(amp.AMP):
    """
    A Proxy AMP protocol that forwards calls to a wrapped
    callRemote-like callable.
    """
    def __init__(self, wrapped, child):
        """
        @param wrapped: A callRemote-like callable that takes an
                        L{amp.Command} as first argument and other
                        optional keyword arguments afterwards.
        @type wrapped: L{callable}.

        @param child: The protocol class of the process pool children.
                      Used to forward only the methods that are actually
                      understood correctly by them.
        @type child: L{amp.AMP}
        """
        amp.AMP.__init__(self)
        self.wrapped = wrapped
        self.child = child

        localCd = set(self._commandDispatch.keys())
        childCd = set(self.child._commandDispatch.keys())
        assert localCd.intersection(childCd) == set([b"StartTLS"]), \
                    "Illegal method overriding in Proxy"

    def locateResponder(self, name):
        """
        This is a custom locator to forward calls to the children
        processes while keeping the ProcessPool a transparent MITM.

        This way of working has a few limitations, the first of which
        is the fact that children won't be able to take advantage of
        any dynamic locator except for the default L{CommandLocator}
        that is based on the _commandDispatch attribute added by the
        metaclass. This limitation might be lifted in the future.
        """
        if name == "StartTLS":
            # This is a special case where the proxy takes precedence
            return amp.AMP.locateResponder(self, "StartTLS")

        # Get the dict of commands from the child AMP implementation.
        cd = self.child._commandDispatch
        if name in cd:
            # If the command is there, then we forward stuff to it.
            commandClass, _responderFunc = cd[name]
            # We need to wrap the doWork function because the wrapping
            # call doesn't pass the command as first argument since it
            # thinks that we are the actual receivers and callable is
            # already the responder while it isn't.
            doWork = lambda **kw: self.wrapped(commandClass, **kw)
            # Now let's call the right function and wrap the result
            # dictionary.
            return self._wrapWithSerialization(doWork, commandClass)
        # of course if the name of the command is not in the child it
        # means that it might be in this class, so fallback to the
        # default behavior of this module.
        return amp.AMP.locateResponder(self, name)




#!/usr/bin/env python
# -*- test-case-name: ampoule -*-

# Copyright (c) 2008 Valentino Volonghi.
# See LICENSE for details.

"""
Distutils/Setuptools installer for AMPoule.
"""

from setuptools import setup

install_requires = ["Twisted[tls]>=17"]

description = """A process pool built on Twisted and AMP."""
long_description = open('README.md').read()

setup(
    name = "ampoule",
    author = "Valentino Volonghi",
    author_email = "dialtone@gmail.com",
    maintainer = "Glyph Lefkowitz",
    maintainer_email = "glyph@twistedmatrix.com",
    description = description,
    description_content_type='text/markdown',
    long_description = long_description,
    long_description_content_type='text/markdown',
    license = "MIT License",
    install_requires=install_requires + ['incremental'],
    url="https://github.com/glyph/ampoule",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: System',
    ],
    packages=["ampoule", "ampoule.test"],
    package_data={'twisted': ['plugins/ampoule_plugin.py']},
    use_incremental=True,
    setup_requires=['incremental'],
    include_package_data = True,
    zip_safe=False
)



from zope.interface import Interface

class IStarter(Interface):
    def startAMPProcess(ampChild, ampParent=None):
        """
        @param ampChild: The AMP protocol spoken by the created child.
        @type ampChild: L{twisted.protocols.amp.AMP}

        @param ampParent: The AMP protocol spoken by the parent.
        @type ampParent: L{twisted.protocols.amp.AMP}
        """

    def startPythonProcess(prot, *args):
        """
        @param prot: a L{protocol.ProcessProtocol} subclass
        @type prot: L{protocol.ProcessProtocol}

        @param args: a tuple of arguments that will be passed to the
                    child process.

        @return: a tuple of the child process and the deferred finished.
                 finished triggers when the subprocess dies for any reason.
        """




#!/usr/bin/env python

# Before using me you need to startup an ampoule process that uses the
# mapreduce child, like this:
# twistd -no ampoule --child=mapreduce.MapReducer

from twisted.internet.protocol import ClientFactory
from twisted.python import filepath as fp
from twisted.internet import reactor, defer
from twisted.protocols import amp

from mapreduce import Map, Reduce
from ampoule.util import mainpoint

import collections

def mymap(in_, out):
    """
    I'm the mapping function executed in the pool.
    """
    aggregation = collections.defaultdict(lambda : 0)

    for line in in_.getContent().splitlines():
        for word in line.split():
            word.strip(",:;.'\"[]{}()*&^%$#@!`~><\\+=-_")
            aggregation[word] += 1

    f = out.open('wb')
    f.write("\n".join("%s %s" % word_frequency
                      for word_frequency in aggregation.items()))
    f.close()


def myreduce(files, out):
    """
    I'm the reducer function applied to a directory in the pool.
    """
    aggregation = collections.defaultdict(lambda : 0)
    for f in files:
        for line in f.getContent().splitlines():
            key, freq = line.split()
            aggregation[key] += int(freq)

    sorted_aggregation = sorted(aggregation.items(),
                                key=lambda t: t[1], reverse=True)
    f = out.open('ab')
    f.write("\n".join("%s %s" % word_frequency for word_frequency in sorted_aggregation))
    f.write('\n')
    f.close()

def map_step(pool, directory, resultdir):
    """
    I define how to walk in the directory that I was given and apply
    the mapper function, then whatever I return is passed to the
    successive steps as first argument.
    """
    outputdir = resultdir.child(directory.basename())
    outputdir.createDirectory()

    l = []
    for filename in directory.children():
        l.append(pool.callRemote(Map, mapper=mymap, filename=filename, outdir=outputdir))

    return defer.DeferredList(l).addCallback(lambda _: outputdir)

def reduce_step(outputdir, pool, resultdir):
    """
    I'm the reduce step and our reduce function assumes that the directory
    is fully processed and that there's a single reduce process per directory.
    """
    return pool.callRemote(Reduce, reducer=myreduce, directory=outputdir)









# Some boilerplate code that is useful to define how to use the
# functions above but nothing more than that.


class AMPFactory(ClientFactory):
    """
    I store variables useful for the protocol connected to the pool.
    """

    def __init__(self, dirs, resultdir, steps):
        self.dirs = dirs
        self.resultdir = resultdir
        self.steps = steps

class ClientAMP(amp.AMP):
    """
    The main orchestration logic in this process, how every function
    in this file is called and in which order and what they do.
    """
    @defer.inlineCallbacks
    def connectionMade(self):
        amp.AMP.connectionMade(self)
        resultdir = self.factory.resultdir
        steps = self.factory.steps
        l = []
        for directory in self.factory.dirs:
            d = steps[0](self, directory, resultdir)
            for step in steps[1:]:
                d = d.addCallback(step, self, resultdir)
            l.append(d)
        yield defer.DeferredList(l)
        reactor.stop()

def process(dirs, resultdir, steps):
    """
    setup twisted and run it with the known parameters.
    """
    c = AMPFactory(dirs, resultdir, steps)
    c.protocol = ClientAMP
    reactor.connectTCP("127.0.0.1", 8901, c)
    reactor.run()

@mainpoint
def main(args):
    """
    The mainpoint, the decorator makes this module resolvable by
    reflect.qual() used by our Function argument type. Here we
    define the steps of our processing and we execute them.
    """
    directory_names = [fp.FilePath(name) for name in args[2:]]
    results = fp.FilePath(args[1])
    results.remove()
    results.makedirs()
    steps = [map_step, reduce_step]
    process(directory_names, results, steps)

